from collections import namedtuple
from enum import Enum
import pickle
import string_util
import os

import util
import citac
import neuralna_mreza as nm
import poravnavanje

RijecInfo = namedtuple('RijecInfo',
                       ['rijec',
                        'glasovi',
                        'interval1',
                        'interval2',
                        'int_maks_1',
                        'ton_maks_1',
                        'int_maks_2',
                        'ton_maks_2'])

Rezultat = namedtuple('Rezultat',
                      ['rijec',
                       'interval1',
                       'int_maks_1',
                       'odst_int_maks_1',
                       'int_oznaka_1',
                       'ton_maks_1',
                       'odst_ton_maks_1',
                       'ton_oznaka_1',
                       'interval2',
                       'int_maks_2',
                       'odst_int_maks_2',
                       'int_oznaka_2',
                       'ton_maks_2',
                       'odst_ton_maks_2',
                       'ton_oznaka_2',
                       'trajanje',
                       'trajanje_oznaka'])


class Detektor:
    def __init__(
            self,
            direktorij,
            pocetno_vrijeme_titla,
            pomak,
            odstupanje_int,
            odstupanje_ton,
            odstupanje_trajanje):
        """

        :param direktorij: Direktorij s podacima za analizu segmenta.
        :param pocetno_vrijeme_titla: Pozicija titla na cijeloj snimci (kako je
            dobiveno iz transkripta).
        :param pomak: Gdje se izdvojeni segment nalazi na cijeloj snimci (sec).
            Ovaj parametar je vazan jer su titlovi oznaceni u odnosu na cijelu
            snimku.
        :param odstupanje_int: Prag odstupanja intenziteta.
        :param odstupanje_ton: Prag odstupanja tona.
        """
        self._direktorij = direktorij
        self._datoteka_intenziteta = 'int_10ms.txt'
        self._datoteka_tona = 'pitch_10ms.txt'
        self._pocetno_vrijeme_titla = pocetno_vrijeme_titla
        self._pomak = pomak
        self.poravnanje = None
        self._odstupanje_int = odstupanje_int
        self._odstupanje_ton = odstupanje_ton
        self._odstupanje_trajanje = odstupanje_trajanje
        self.vrijeme = self
        self._rezultat_analize = []

    def inicijalizacija(self, tekst_titla, interval_titla):
        """
        Inicijalizacija pocetnih vrijednosti.

        :param tekst_titla:
        :param interval_titla:
        :return: Lista RijecInfo objekata.
        """
        tabela = []
        pauze = self.__intervali_pauza()
        rijeci = tekst_titla.split(' ')
        tekst = tekst_titla.replace(' ', '').lower()
        interval_zvuka = interval_titla[1] - interval_titla[0]
        nastavak = interval_titla[0]

        for interval_rijeci in rijeci:
            rijec = util.PripremaTeksta(interval_rijeci) \
                .obradjen_tekst().lower()
            postotak_teksta = len(rijec) / len(tekst) * 100
            utroseni_dio_zvuka = interval_zvuka * postotak_teksta / 100

            od, do = round(nastavak, 2), round(nastavak + utroseni_dio_zvuka, 2)
            info = (rijec, od, do)
            interval_rijeci = Detektor.__uklopi_pauze(pauze, info)
            if interval_rijeci != info:
                nastavak = interval_rijeci[2]
            else:
                nastavak += utroseni_dio_zvuka

            tabela += [RijecInfo(rijec=interval_rijeci[0],
                                 glasovi=[],
                                 interval1=interval_rijeci[1:],
                                 interval2=None,
                                 int_maks_1=0,
                                 ton_maks_1=0,
                                 int_maks_2=0,
                                 ton_maks_2=0)]

        return tabela

    def __interval_titla(
            self,
            pozicija_pocetnog_titla,
            pozicija_segmenta,
            vrijeme_od,
            vrijeme_do):
        """
        S obzirom da je segment zvuka za analizu izvucen iz vece snimke ova
        funkcija izracunava poziciju titla na izdvojenom segmentu na osnovu
        njegove pozicije na cijeloj (vecoj) snimci. Podatke o poziciji titla
        koji dolaze preko teleteksta NE modificiramo u izvornoj (TS) datoteci!

        :param pozicija_pocetnog_titla: Pozicija titla na cijeloj snimci.
        :param pozicija_segmenta: Pozicija izdvojenog segmenta zvuka na cijeloj
                                    snimci (pomak).
        :param vrijeme_od: Pocetno vrijeme titla (iz TS datoteke).
        :param vrijeme_do: Zavrsno vrijeme titla (iz TS datoteke).
        :return: Preracunati interval titla u odnosu na segment snimke.
        """

        def pomak_titla(pozicija_titla: tuple):
            intenziteti = citac.vrijeme_vrijednost_dat(
                    self._direktorij,
                    self._datoteka_intenziteta)
            iznad_praga = [p for p in intenziteti if p[1] > 60]
            pocetak_zvuka = iznad_praga[0][0]
            return (pozicija_titla[0] * 60 + pozicija_titla[1]
                    - pozicija_segmenta - pocetak_zvuka)

        pomak = pomak_titla(pozicija_pocetnog_titla)
        pocetak = (util.sekunde(vrijeme_od[0], vrijeme_od[1]) - pomak
                   - pozicija_segmenta)
        zavrsetak = (util.sekunde(vrijeme_do[0], vrijeme_do[1]) - pomak
                     - pozicija_segmenta)
        # pozicija = (round(pocetak, 2), round(zavrsetak, 2))
        return round(pocetak, 2), round(zavrsetak, 2)

    def __intervali_pauza(self):
        """
        Daje pozicije zvuka s intenzitetom ispod zadanog praga (tisina).

        :return: Lista parova (od, do).
        """
        intenziteti = citac.vrijeme_vrijednost_dat(
                self._direktorij,
                self._datoteka_intenziteta)
        ispod_praga = [p for p in intenziteti if p[1] < 60]  # <60 dB
        pauze = []

        interval = []
        prethodno_vrijeme = round(ispod_praga[0][0], 2)
        for i in range(1, len(ispod_praga)):
            vrijeme = round(ispod_praga[i][0], 2)
            if round(abs(vrijeme - prethodno_vrijeme), 2) <= 0.01:
                interval += [vrijeme]
                prethodno_vrijeme = round(ispod_praga[i][0], 2)
            else:
                if len(interval) > 1 \
                        and interval[-1] - interval[0] >= 0.077:
                    pauze += [(interval[0], interval[-1])]
                interval = []
                prethodno_vrijeme = round(ispod_praga[i][0], 2)

        return pauze

    @staticmethod
    def __uklopi_pauze(pauze, rijec_info: tuple):
        """
        Dodavanje vremena pauze u trajanje rijeci (zbog preciznijeg poravnanja).

        :param pauze: Podaci o pozicijama pauza.
        :param rijec_info: Podaci o rijeci za koju treba uklopiti pauze.
        :return: ntorka (rijec, od, do)
        """

        def izmedju(n, interval):
            return interval[0] <= n <= interval[1]

        def unutar(interval_1, interval_2):
            return izmedju(interval_1[0], interval_2) \
                   and izmedju(interval_1[1], interval_2)

        for p in pauze:
            if unutar(p, rijec_info[1:]):
                return (rijec_info[0],
                        rijec_info[1],
                        round(rijec_info[2] + (p[1] - p[0]), 2))

        return rijec_info

    def __postavi_zvucne_podatke(self, rijeci):
        """
        Postavljanje podataka o tonu i intenzitetu za segmente zvuka kojima
        odgovaraju rijeci.

        :param rijeci: Popis objekata RijecInfo.
        :return: Popis objekata RijecInfo.
        """
        popis = []
        for rijec_info in rijeci:
            intenziteti_1 = [n for (v, n) in citac.vrijeme_vrijednost_dat(
                    self._direktorij,
                    'int_10ms.txt',
                    interval=rijec_info.interval1)]
            tonovi_1 = [n for (v, n) in citac.vrijeme_vrijednost_dat(
                    self._direktorij,
                    'pitch_10ms.txt',
                    interval=rijec_info.interval1)]

            intenziteti_2 = [n for (v, n) in citac.vrijeme_vrijednost_dat(
                    self._direktorij,
                    'int_10ms.txt',
                    interval=rijec_info.interval2)]
            tonovi_2 = [n for (v, n) in citac.vrijeme_vrijednost_dat(
                    self._direktorij,
                    'pitch_10ms.txt',
                    interval=rijec_info.interval2)]

            int_maks_1 = max(intenziteti_1) \
                if intenziteti_1 not in (None, []) \
                else None
            ton_maks_1 = max(tonovi_1) \
                if tonovi_1 not in (None, []) else None
            int_maks_2 = max(intenziteti_2) \
                if intenziteti_2 not in (None, []) else None
            ton_maks_2 = max(tonovi_2) \
                if tonovi_2 not in (None, []) else None
            popis += [RijecInfo(
                    rijec=rijec_info.rijec,
                    glasovi=rijec_info.glasovi,
                    interval1=rijec_info.interval1,
                    interval2=rijec_info.interval2,
                    int_maks_1=int_maks_1,
                    ton_maks_1=ton_maks_1,
                    int_maks_2=int_maks_2,
                    ton_maks_2=ton_maks_2)]

        return popis

    @staticmethod
    def __izdvoji_glasove_rijeci(tabela_glasova, poravnanje):
        """
        Za intervale poravnanja izdvaja klasificirane glasove (da se moze
        mjeriti trajanje).

        :param tabela_glasova:
        :param poravnanje:
        :return: Lista mapa {od, do, grupa_glasova}.
        """

        class Stanje(Enum):
            PRESKOCI = 0
            DODAJ = 1

        glasovi_rijeci = []
        for rijec_info in poravnanje:
            stanje = Stanje.PRESKOCI
            glasovi = []
            for glas_info in tabela_glasova:
                if stanje == Stanje.DODAJ:
                    if glas_info['do'] <= rijec_info.interval1[1]:
                        glasovi += [glas_info['grupa_glasova']]
                    else:
                        glasovi_rijeci += [glasovi]
                        # prekidamo dodavanje glasova i idemo na iducu rijec
                        break
                else:
                    if glas_info['od'] >= rijec_info.interval1[0]:
                        stanje = Stanje.DODAJ
                        glasovi += [glas_info['grupa_glasova']]

        return glasovi_rijeci

    def __poravnanje(
            self,
            titl,
            vrijeme_titla_od,
            vrijeme_titla_do):
        """
        Postavljanje poravnanja na osnovu klasificiranih glasova.

        :param titl: Tekst transkripta.
        :param vrijeme_titla_od: Pocetak titla.
        :param vrijeme_titla_do: Zavrsetak titla.
        :return: Lista RijecInfo objekata.
        """
        pozicija_titla = self.__interval_titla(
                self._pocetno_vrijeme_titla,
                self._pomak,
                vrijeme_titla_od,
                vrijeme_titla_do)

        print('INTERVAL TITLA: ', pozicija_titla)
        glasovi = klasifikacija(self._direktorij)
        tabela_glasova = poravnavanje.napravi_tabelu_glasova(
                glasovi,
                pozicija_titla[0] - 0.2,
                pozicija_titla[1] + 0.2)
        # print('=> Glasovi:', tabela_glasova)
        priprema_teksta_obj = util.PripremaTeksta(titl)
        uredjen_tekst = priprema_teksta_obj.obradjen_tekst()
        self.poravnanje, _ = poravnavanje.poravnaj_slova_s_glasovima(
                uredjen_tekst.lower(),
                tabela_glasova)
        print(string_util.String(self.vrijeme, titl, pozicija_titla))
        intervali_rijeci = poravnavanje.intervali_rijeci(
                priprema_teksta_obj,
                self.poravnanje)
        glasovi_rijeci = Detektor.__izdvoji_glasove_rijeci(tabela_glasova,
                                                           self.poravnanje)

        # postavi intervale rijeci
        rijeci = []
        for i, info in enumerate(self.poravnanje):
            rijeci += [RijecInfo(
                    rijec=info.rijec,
                    glasovi=glasovi_rijeci[i]
                    if i < len(glasovi_rijeci) else [],
                    interval1=info.interval1,
                    interval2=(intervali_rijeci[info.rijec]
                               if info.rijec in intervali_rijeci else (0, 0)),
                    int_maks_1=0,
                    ton_maks_1=0,
                    int_maks_2=0,
                    ton_maks_2=0)]

        return self.__postavi_zvucne_podatke(rijeci)

    def __prosjek_intenziteta_i_tona(self, interval):
        """
        Daje podatke o prosjecnom intenzitetu i tonu za zadani interval zvuka.

        :param interval: Interval zvuka.
        :return: Par (prosjecni intenzitet, prosjecni ton)
        """
        intz = citac.vrijeme_vrijednost_dat(
                self._direktorij,
                self._datoteka_intenziteta)
        ton = citac.vrijeme_vrijednost_dat(
                self._direktorij,
                self._datoteka_tona)
        vr_int = [n for (v, n) in intz if interval[0] <= v <= interval[1]]
        vr_ton = [n for (v, n) in ton if interval[0] <= v <= interval[1]]
        pr_int = sum(vr_int) / len(vr_int) if len(vr_int) > 0 else 0
        pr_ton = sum(vr_ton) / len(vr_ton) if len(vr_ton) > 0 else 0
        return round(pr_int, 2), round(pr_ton, 2)

    @staticmethod
    def __trajanja(rijeci):
        """
        Izdvoji najduzi niz vokala i izracunaj odstupanje njegove duljine od
        ostalih nizova vokala unutar titla ili prozora.

        :param rijeci:
        :return:
        """
        duljine_niza_vokala = []
        for r in rijeci:
            duljine_vokala_za_rijec = []
            for grupa_glasova in r.glasovi:
                if grupa_glasova[0] in 'ieaou':
                    duljine_vokala_za_rijec += [len(grupa_glasova)]
            duljine_niza_vokala.append(duljine_vokala_za_rijec)

        prosjek_trajanja_rijeci = [sum(v) / len(v) if len(v) > 0 else -1
                                   for v in duljine_niza_vokala]
        prosjek_trajanja_za_titl = sum(prosjek_trajanja_rijeci) / len(rijeci)
        return prosjek_trajanja_za_titl, prosjek_trajanja_rijeci

    def __analiza(self, titl, od, do):
        """
        Analiza zvuka za svaku rijec gdje je interval rijeci dobiven prethodnim
        poravnanjem.

        :param titl: Tekst transkripta.
        :param od: Pocetak titla.
        :param do: Zavrsetak titla.
        :return: self
        """
        rijeci = self.__poravnanje(titl, od, do)  # lista RijecInfo objekata
        # print('==> Tabela:', rijeci)

        # odredi prosjek intenziteta i tona za cijeli titl
        pr_int, pr_ton = self.__prosjek_intenziteta_i_tona(
                (rijeci[0].interval1[0], rijeci[-1].interval1[1]))

        # odredi prosjek trajanja
        pr_trajanja_titl, pr_trajanja_rijeci = Detektor.__trajanja(rijeci)
        print(f'PROSJEK INT.: {pr_int:<10}PROSJEK TON: {pr_ton:<10}'
              f'PROSJEK TRAJANJE: {str(round(pr_trajanja_titl, 2)):<10}')

        self._rezultat_analize = []
        for i in range(len(rijeci)):
            int_maks_1 = rijeci[i].int_maks_1
            ton_maks_1 = rijeci[i].ton_maks_1
            int_maks_2 = rijeci[i].int_maks_2
            ton_maks_2 = rijeci[i].ton_maks_2

            if int_maks_1 is None or ton_maks_1 is None:
                odst_int_maks_1 = 0
                odst_ton_maks_1 = 0
            else:
                odst_int_maks_1 = round((int_maks_1 / pr_int - 1) * 100, 2)
                odst_ton_maks_1 = round((ton_maks_1 / pr_ton - 1) * 100, 2)

            if int_maks_2 is None or ton_maks_2 is None:
                odst_int_maks_2 = 0
                odst_ton_maks_2 = 0
            else:
                odst_int_maks_2 = round((int_maks_2 / pr_int - 1) * 100, 2)
                odst_ton_maks_2 = round((ton_maks_2 / pr_ton - 1) * 100, 2)

            int_oznaka_1 = ''
            if odst_int_maks_1 > self._odstupanje_int:
                int_oznaka_1 = '*'

            int_oznaka_2 = ''
            if odst_int_maks_2 > self._odstupanje_int:
                int_oznaka_2 = '*'

            ton_oznaka_1 = ''
            if odst_ton_maks_1 > self._odstupanje_ton:
                ton_oznaka_1 = '*'

            ton_oznaka_2 = ''
            if odst_ton_maks_2 > self._odstupanje_ton:
                ton_oznaka_2 = '*'

            trajanje = round((pr_trajanja_rijeci[i] - pr_trajanja_titl)
                             / pr_trajanja_titl * 100, 2)
            trajanje_oznaka = ('*'
                               if trajanje > self._odstupanje_trajanje else '')

            rezultat = Rezultat(
                    rijec=rijeci[i].rijec,
                    interval1=rijeci[i].interval1,
                    int_maks_1=int_maks_1,
                    odst_int_maks_1=odst_int_maks_1,
                    int_oznaka_1=int_oznaka_1,
                    ton_maks_1=ton_maks_1,
                    odst_ton_maks_1=odst_ton_maks_1,
                    ton_oznaka_1=ton_oznaka_1,
                    interval2=rijeci[i].interval2,
                    int_maks_2=int_maks_2,
                    odst_int_maks_2=odst_int_maks_2,
                    int_oznaka_2=int_oznaka_2,
                    ton_maks_2=ton_maks_2,
                    odst_ton_maks_2=odst_ton_maks_2,
                    ton_oznaka_2=ton_oznaka_2,
                    trajanje=trajanje,
                    trajanje_oznaka=trajanje_oznaka)
            self._rezultat_analize += [rezultat]

        return self

    def __formatiraj_rezultat(self):
        """
        Prikaz rezultata tabularno i SSML.

        :return: None
        """
        ssml = '<speak>'
        print(f'{"RIJEC":15}'
              f'{"INTERVAL":15}'
              f'{"MI (dB)":13}'
              f'{"Odst. MI":12}'
              f'{"Ist. int.":12}'
              f'{"MT (Hz)":13}'
              f'{"Odst. MT":12}'
              f'{"Ist. ton":12}'
              f'{"Ist. trajanje":12}')
        for rijec_info in self._rezultat_analize:
            print(f'{rijec_info.rijec:15}'
                  f'{str(rijec_info.interval1):15}'
                  f'{str(rijec_info.int_maks_1):13}'
                  f'{str(rijec_info.odst_int_maks_1) + " %":12}'
                  f'{str(rijec_info.int_oznaka_1):^12}'
                  f'{str(rijec_info.ton_maks_1):13}'
                  f'{str(rijec_info.odst_ton_maks_1) + " %":12}'
                  f'{str(rijec_info.ton_oznaka_1):^12}'
                  f'{str(rijec_info.trajanje_oznaka):^12}')

            dodaj_prozodiju = False
            prozodija = '<prosody '
            if rijec_info.int_oznaka_1 == '*':
                # 6 je maksimum (prema pravilima SSMLa)
                db = round(6 * (rijec_info.odst_int_maks_1 / 100), 2)
                prozodija += 'volume=+' + str(db) + 'dB '
                dodaj_prozodiju = True

            if rijec_info.ton_oznaka_1 == '*':
                prozodija += 'pitch=+' + str(rijec_info.odst_ton_maks_1) + '%'
                dodaj_prozodiju = True

            prozodija += '> ' + rijec_info.rijec + ' </prosody> '

            if dodaj_prozodiju:
                ssml += prozodija
            else:
                ssml += rijec_info.rijec + ' '

        ssml += '</speak>'
        print('\nSSML:', ssml, '\n')

    def prikazi(self):
        """
        Glavna metoda za prikaz rezultata automatske detekcije naglasenih
        rijeci.

        :return: None
        """
        titlovi = citac.Titlovi(self._direktorij)
        popis_titlova = titlovi.ucitaj()
        for titl in popis_titlova:
            rezultat = self.__analiza(titl.titl, titl.vrijeme_od,
                                      titl.vrijeme_do)
            rezultat.__formatiraj_rezultat()

# ==============================================================================


def klasifikacija(putanja):
    """
    Ovo se koristi samo ako treba nanovo klasificirati glasove. Ako datoteka
    s klasificiranim glasovima vec postoji onda ova funkcija vraca popis
    tih glasova, inace se glasovi nanovo klasificiraju i taj se popis glasova
    vraca.
    :param putanja: Putanja gdje treba smjestiti datoteku s klasificiranim
                    glasovima.
    :return: Klasificirane glasove
    """
    datoteka = util.konfig('datoteka', 'dir') + putanja + '/glasovi.pyobj'
    if os.path.exists(datoteka):
        with open(datoteka, 'rb') as f:
            glasovi = pickle.load(f)
            return glasovi
    else:
        glasovi = nm.klasificiraj_glasove(putanja)
        with open(datoteka, 'wb') as f:
            pickle.dump(glasovi, f, pickle.HIGHEST_PROTOCOL)
        return glasovi


def proba():
    odst_int = 20  # odstupanje intenziteta vece od ovog postotka je nagl. rijec
    odst_ton = 25  # odstupanje tona vece od ovog postotka je nagl. rijec
    odst_trajanje = 250  # odstupanje trajanja od ovog postotka je nagl. rijec

    info = [
        # ('/detekcija/test', (0, 0), 0),

        ('/detekcija/proba_1_m1', (7, 39.44), 459),
        # ('/detekcija/proba_2_m1', (7, 48.44), 467.7),
        # ('/detekcija/proba_5_m2', (36, 7.68), 2166.85),
        # ('/detekcija/proba_12_m4', (40, 20.04), 2418.56),
        # ('/detekcija/proba_3_z1', (7, 27.32), 446.7),
        # ('/detekcija/proba_8_z4', (10, 13.08), 612.5),
        # ('/detekcija/proba_10_z5', (12, 0), 719.2),
        # ('/detekcija/proba_11_z5', (11, 42.36), 701.56),
    ]

    for p in info:
        print('=' * 120, '\n***** DIREKTORIJ:', p[0])
        Detektor(*p, odst_int, odst_ton, odst_trajanje).prikazi()


if __name__ == '__main__':
    proba()
