import statistics

import podjela_na_slogove


def izdvoji_rijeci(
        ispravni_intervali,
        tabela_glasova,
        tg_poravnanje,
        intenzitet,
        ton,
        obrada_teksta):
    """

    :param tabela_glasova:
    :param ton:
    :param intenzitet:
    :param ispravni_intervali: Intervali rijeci utvrdjeni slusanjem.
    :param tg_poravnanje: Poravnanje teksta s glasovima (vidi funkciju).
    :param obrada_teksta: Objekt tipa PripremaTeksta.

    :return: Niz mapa gdje svaka mapa predstavlja informacije o jednoj
    rijeci, formata [{'rijec':..., 'od':..., 'do':..., ...}, {...}, ...].
    """

    def _intervali_rijeci():
        """
        Razdvaja popis glasova prema granicama rijeci (oznacenim s ' ').

        :return: Niz s informacijama o poziciji glasova; svaka podlista je
        jedna rijec. Primjer rijeci 'su':
            [..., [('s', (21.9523, 22.06)), ('u', (22.07, 22.086))], ...]
        """
        izvorni_niz = obrada_teksta.izvorni_niz(tg_poravnanje)
        intervali_rijeci = []
        jedan_interval = []
        for e in izvorni_niz:
            if e == ' ':  # zavrsava rijec?
                intervali_rijeci += [jedan_interval]
                jedan_interval = []
            else:
                jedan_interval += [e]

        intervali_rijeci += [jedan_interval]  # dodaj ostatak
        return intervali_rijeci

    def _popis_rijeci(intervali_rijeci):
        """
        Daje listu rijeci s informacijom o vremenskoj poziciji.

        :param intervali_rijeci: (gornja funkcija)
        :return: Niz formata [(rijec, od, do), ...].
        """
        popis_rijeci = []
        for interval in intervali_rijeci:
            pocetno_vrijeme, zavrsno_vrijeme = None, None
            rijec = ''
            for e in interval:
                if isinstance(e, tuple):
                    if pocetno_vrijeme is None:
                        pocetno_vrijeme = e[1][0]
                    zavrsno_vrijeme = e[1][1]
                    rijec += e[0]
                else:
                    rijec += e

            if rijec != '':
                popis_rijeci += [(rijec, pocetno_vrijeme, zavrsno_vrijeme)]

        return popis_rijeci

    def _tabela(popis_rijeci):
        """
        Formira tabelu koja sadrzi sve potrebne podatke.
        :param popis_rijeci: (vidi poziv)
        :return: Niz mapa.
        """
        tabela = []
        for i in range(len(popis_rijeci)):
            if i >= len(ispravni_intervali):
                break
            rijec = popis_rijeci[i][0]
            od = popis_rijeci[i][1]
            do = popis_rijeci[i][2]
            ispravno_od = ispravni_intervali[i][1]
            ispravno_do = ispravni_intervali[i][2]
            odst_od = (popis_rijeci[i][1] - ispravni_intervali[i][1]) * 1000
            odst_do = (popis_rijeci[i][2] - ispravni_intervali[i][2]) * 1000
            slogovi = podjela_na_slogove.Slogovi(rijec).podjela()
            br_slogova = len(slogovi)

            # intenzitet
            int_segment = list(filter(lambda v: od <= v[0] <= do, intenzitet))
            int_min = -1
            int_maks = -1
            int_prosjek = -1
            if len(int_segment) > 0:
                int_min = round(min([v[1] for v in int_segment]), 2)
                int_maks = round(max([v[1] for v in int_segment]), 2)
                int_prosjek = round(statistics.mean([v[1]
                                                     for v in int_segment]), 2)

            # ton
            ton_segment = list(filter(lambda v: od <= v[0] <= do, ton))
            ton_min = -1
            ton_maks = -1
            ton_prosjek = -1
            if len(ton_segment) > 0:
                ton_min = round(min([v[1] for v in ton_segment]), 2)
                ton_maks = round(max([v[1] for v in ton_segment]), 2)
                ton_prosjek = round(statistics.mean([v[1]
                                                     for v in ton_segment]), 2)

            # trajanje: za svaku rijec izdvajamo najdulji niz vokala
            interval_rijeci = od, do
            segment_glasova = list(
                    filter(lambda g: g['od'] >= interval_rijeci[0]
                                     and g['do'] <= interval_rijeci[1]
                                     and g['grupa_glasova'][0] in 'ieaou',
                           tabela_glasova))
            if len(segment_glasova) > 0:
                sortirani_segment_glasova = \
                    sorted(segment_glasova,
                           key=lambda e: len(e['grupa_glasova']))
                duljina_vokala = \
                    len(sortirani_segment_glasova[-1]['grupa_glasova'])
            else:
                duljina_vokala = 0

            redak = {
                'rijec': rijec,
                'od': od,
                'do': do,
                'ispravno_od': ispravno_od,
                'ispravno_do': ispravno_do,
                'odstupanje_od': odst_od,
                'odstupanje_do': odst_do,
                'broj_slogova': br_slogova,
                'slogovi': slogovi,
                'intenzitet': (int_min, int_maks, int_prosjek),
                'ton': (ton_min, ton_maks, ton_prosjek),
                'trajanje': duljina_vokala,

                # ovo se postavlja kasnije kod analize prozodije
                'istaknut_intenzitet': None,
                'istaknut_ton': None,
                'istaknuto_trajanje': None,
            }

            tabela += [redak]

        return tabela

    intervali = _intervali_rijeci()
    rijeci = _popis_rijeci(intervali)
    return _tabela(rijeci)


def analiza_tempa(tabela, min_br_slogova):
    """
    Iz popisa rijeci izdvaja rijeci u grupe od minimalno min_br_slogova.

    :param min_br_slogova: Minimalni broj slogova po grupi.
    :param tabela: Tabela s rijecima.

    :return: Lista redova tabele (mapa) grupiranih tako da sadrze najmanje
    min_br_slogova. FORMAT: [[{redak}, {redak}, ...], [...], ...].
    """
    grupe = []
    grupa_redova = []
    br_slogova = 0
    for redak in tabela:
        grupa_redova += [redak]
        br_slogova += redak['broj_slogova']
        if br_slogova >= min_br_slogova:  # pripremi za novu grupu
            grupe += [grupa_redova]
            grupa_redova = []
            br_slogova = 0

    if len(grupa_redova) > 0:
        grupe += [grupa_redova]  # dodaj ostatak

    return grupe


def analiziraj_prozodiju(
        tabela_rijeci,
        tabela_glasova,
        odstupanje_ton,
        odstupanje_int,
        odstupanje_trajanje,
        velicina_prozora):
    """
    Ova funkcija postavlja informacije o intenzitetu, tonu i trajanju u tabelu
    rijeci.

    :param tabela_glasova:
    :param velicina_prozora: Koliko rijeci uzimamo kao kontekst za detekciju
                             trajanja.
    :param tabela_rijeci: Popis rijeci sa svim informacijama.
    :param odstupanje_ton: Postotak odstupanja od okolnih vrijednosti koji se
                            smatra isticanjem.
    :param odstupanje_int: Isto kao za ton.
    :param odstupanje_trajanje: Isto kao za ton.
    :return: None
    """

    def _istaknut_ton():
        for i in range(len(tabela_rijeci)):
            redak = tabela_rijeci[i]
            ton_maks = redak['ton'][1]
            if 0 < i < len(tabela_rijeci) - 1:
                ton_prethodni = tabela_rijeci[i - 1]['ton'][1]
                ton_iduci = tabela_rijeci[i + 1]['ton'][1]
                if ton_maks >= ton_prethodni + ton_prethodni * odstupanje_ton \
                        and ton_maks >= ton_iduci + ton_iduci * odstupanje_ton:
                    tabela_rijeci[i]['istaknut_ton'] = True

    def _istaknut_intenzitet():
        for i in range(len(tabela_rijeci)):
            redak = tabela_rijeci[i]
            int_maks = redak['intenzitet'][1]
            if 0 < i < len(tabela_rijeci) - 1:
                int_prethodni = tabela_rijeci[i - 1]['intenzitet'][1]
                int_iduci = tabela_rijeci[i + 1]['intenzitet'][1]
                if int_maks >= int_prethodni + int_prethodni * odstupanje_int \
                        and int_maks >= int_iduci + int_iduci * odstupanje_int:
                    tabela_rijeci[i]['istaknut_intenzitet'] = True

    def _najduzi_segment_vokala(od, do):
        """
        Vraca segment vokala za zadani interval glasova.

        :param od:
        :param do:
        :return: Lista grupa vokala zadanog segmenta, npr.
                ['aaaaea', 'iieieeaa', ...].
        """
        vokali_rijeci = []
        for redak in tabela_glasova:
            if redak['od'] >= od \
                    and redak['do'] <= do \
                    and redak['grupa_glasova'][0] in 'ieaou':
                vokali_rijeci += [redak['grupa_glasova']]

        if len(vokali_rijeci) > 0:
            return max(vokali_rijeci, key=len)
        else:
            return ''

    def _iznad_praga(prozor, rijec_od, rijec_do, indeks_rijeci):
        najduzi_trenutni = _najduzi_segment_vokala(rijec_od, rijec_do)
        segment_prozora = tabela_rijeci[prozor[0]:indeks_rijeci] \
                          + tabela_rijeci[indeks_rijeci + 1:prozor[1] + 1]
        ostali = []
        for rijec in segment_prozora:
            od = rijec['od']
            do = rijec['do']
            ostali += [_najduzi_segment_vokala(od, do)]

        najduzi_od_ostalih = max(ostali, key=len)
        if len(najduzi_trenutni) > 0:
            if len(najduzi_od_ostalih) / len(najduzi_trenutni) \
                    < odstupanje_trajanje:
                return True
            else:
                return False
        else:
            return False

    def _istaknuto_trajanje():
        """
        Za svaku rijec odredi je li istaknuto trajanje.

        :return: None
        """
        prozor = 0, velicina_prozora - 1

        # sve unutar prozora uzimamo kao kontekst za analizu trenutne rijeci
        for i, redak in enumerate(tabela_rijeci):
            if i > prozor[1]:  # dosli smo do kraja prozora, pocni ga spustati
                prozor = prozor[0] + 1, prozor[1] + 1

            # Prvo uzmemo najdulji segment vokala trenutne rijeci, a zatim
            # najdulji segment ostalih rijeci unutar prozora.
            if _iznad_praga(prozor, redak['od'], redak['do'], i):
                tabela_rijeci[i]['istaknuto_trajanje'] = True

    _istaknut_ton()
    _istaknut_intenzitet()
    _istaknuto_trajanje()
