from sklearn.neural_network import MLPClassifier
import pickle
import os
import re

import citac
import util


def skalirano_01(ltas: map):
    """
    * Skaliranje vrijednosti na interval [0, 1] (potrebno za neuralnu mrezu)
    * Podesavanje vrijednosti na odredjeni intenzitet (tako da ne ovisimo o
    razlikama u intenzitetu)
    
    :param ltas: mapa gdje je kljuc vrijeme, a vrijednost LTAS (LTAS je 
    lista koja se nalazi u drugoj listi)
    :return: Mapa gdje je kljuc naziv glasa, a vrijednost LTAS za taj glas.
    Vrijednost je u stvari lista koja sadrzi drugu listu (zbog treniranja
    i upita za mrezu, ali to je vjerojatno zbog toga sto se mrezi moze
    zadati niz vrijednosti koje treba klasificirati, ne samo jednu kako sad
    radim).
    """

    def _skalirano_po_intenzitetu(ltas_vr):
        najveci = max(ltas_vr)
        granica = 30
        promjena = (najveci - granica) / najveci
        rezultat = list(map(lambda v: v - v * promjena, ltas_vr))
        return rezultat

    def _skalirana_vrijednost_na_01(ltas_vr):
        p = _skalirano_po_intenzitetu(ltas_vr)
        vr_min = min(p)
        vr_max = max(p)

        # algoritam "feature scaling"
        return list(map(lambda x: (x - vr_min) / (vr_max - vr_min), p))

    def _skaliran_glas_na_01(ltas_liste: list):
        return list(
            map(lambda ltas_glasa: _skalirana_vrijednost_na_01(ltas_glasa),
                ltas_liste))

    skalirano = {}
    for naziv in ltas:
        skalirano[naziv] = _skaliran_glas_na_01(ltas[naziv])

    return skalirano


def treniraj():
    """
    Korak 1: Formiraj sljedecu strukturu:
        ([LTAS-1, LTAS-2, ..., LTAS-n],
         [GLAS-1, GLAS-2, ..., GLAS-n])
        
    gdje je LTAS-k lista LTAS vrijednosti, a GLAS-k glas koji odgovara listi
    LTAS-k. Na primjer:
        ([[0.3, 0.1, ...], [0.7, 0.2, ...], ...], ['n', 'b', ...])
        
    Prvi element para je lista drugih lista koje sadrze LTAS vrijednosti
    nekog glasa, a drugi element para je taj glas (na poziciji na kojoj se
    nalazi njegova LTAS lista).
    Svaka LTAS lista ce, pored ostalih vrijednosti, sadrzavati 1 kao najvecu
    vrijednost i 0 kao najmanju.

    Korak 2: Treniraj NM koristeci prethodno napravljene kategorije.
    
    :return: (vidi gore)
    """

    def _serijaliziraj(obj):
        with open(util.konfig('datoteka', 'kategorije'), 'wb') as f:
            pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

    direktorij = util.konfig('datoteka', 'tset')
    ltas = []
    nove_kategorije = []
    print(util.konfig('datoteka', 'tset'))
    oznacene_ltas_datoteke = list(os.walk(util.konfig('datoteka', 'tset')))[0][
        2]
    for dat in oznacene_ltas_datoteke:  # ucitavamo iz svih datoteka
        print('uzorak:', dat)
        tr_set = citac.PraatLTASDatoteka(direktorij + dat)
        tr_set.ucitaj()
        skalirano = skalirano_01(tr_set.vrijednosti())

        for k in skalirano:
            ltas += skalirano[k]
            nove_kategorije += [k] * len(
                skalirano[k])  # jedan glas moze imati vise LTAS lista

    _serijaliziraj((ltas, nove_kategorije))
    print('treniranje ...')
    treniraj_mrezu(util.konfig('datoteka', 'ann'), (ltas, nove_kategorije))


def treniraj_mrezu(ime_ann_datoteke, ltas_kategorije):
    """
    Ovo je format podataka za NM:
        ltas = ([23., 14., ...], [18., 72., ...])   vrijednosti (LTAS)
        kategorije = ['A', 'B']             kategorije za svaku podlistu u ltas

    :param ime_ann_datoteke
    :param ltas_kategorije
    :return: mreza istrenirana sa zadanim LTAS datotekama
    """
    mreza = MLPClassifier(
            solver='lbfgs',  # preporucuje se za manju kolicinu podataka
            alpha=1e-5,      # zadano (default) za korekciju
            hidden_layer_sizes=(120, 120, 160),
            random_state=1)  # za generator slucajnih brojeva

    # izgleda da se mreza mora trenirati odjednom sa svim podacima
    # ltas, kategorije = kategorije_glasova(oznacene_ltas_datoteke)
    ltas, kategorije = ltas_kategorije
    mreza.fit(ltas, kategorije)  # treniranje

    # serijaliziraj treniranu mrezu
    with open(ime_ann_datoteke, 'wb') as f:
        pickle.dump(mreza, f, pickle.HIGHEST_PROTOCOL)

    return mreza


def klasificiraj(direktorij, ltas_datoteka, int_datoteka, izlazna_datoteka):
    """
    Trazimo od trenirane NM da klasificira glasove. Klasificirani glasovi
    se spremaju u izlaznu datoteku. NM je serijalizirani objekt prethodno
    istrenirane mreze.

    :param direktorij: ...
    :param ltas_datoteka: LTAS datoteka neoznacene snimke govora
    :param int_datoteka:
    :param izlazna_datoteka: ...
    :return: Lista ntorki (slovo, vrijeme)
    """

    ulaz = citac.PraatLTASDatoteka(rf'{direktorij}\{ltas_datoteka}')
    ulaz.ucitaj()
    vrijednosti = skalirano_01(ulaz.vrijednosti())

    try:
        with open(util.konfig('datoteka', 'ann'), 'rb') as f:
            mreza = pickle.load(f)

        popis_glasova = []
        with open(rf'{direktorij}\{izlazna_datoteka}', 'w',
                  encoding='utf8') as f:
            # vrijeme_glasa je u stvari naziv koji oznacava vrijeme kao 5_23
            for vrijeme_glasa in vrijednosti:
                vrijeme = round(
                    float(vrijeme_glasa.replace('_', '.')), 4)

                # klasificiraj skalirani glas (rezultat je tipa numpy.ndarray)
                klasificiran_glas = mreza.predict(vrijednosti[vrijeme_glasa])
                slovo = util.slovo_glasa(klasificiran_glas[0])

                # ovdje je vrijeme u formatu x.yz
                popis_glasova += [(slovo, vrijeme)]
                f.write(slovo + ' (' + vrijeme_glasa + ')\n')

        return popis_glasova
    except Exception as e:
        print(e)


def frekv(niz):
    m = {}
    for e in niz:
        if e in m:
            m[e] += 1
        else:
            m[e] = 1

    for e in m:
        m[e] = round(m[e] / len(niz) * 100, 2)

    return m


def glas_pulsa(puls, a, b):
    """
    Ovi glasovi su uzeti iz pulseva.
    :param puls:
    :param a:
    :param b:
    :return: string (ostalih) glasova iz pulseva
    """

    c_glas = 0
    c_vrijeme = 1

    g_ostali_glasovi = [(g[c_glas], g[c_vrijeme]) for g in puls if
                        g[c_glas] not in 'ieaoušđžčcsz']
    r = []
    od = None
    for g in g_ostali_glasovi:
        if a < g[c_vrijeme] < b:
            r += [g]
            if od is None:
                od = g[c_vrijeme]

    if len(r) > 0:
        do = r[-1][c_vrijeme]
    else:
        do = None

    popis_glasova = list(map(lambda e: e[c_glas], r))
    return ''.join(popis_glasova), od, do


def postotak_glasova(niz: str):
    """
    Koliki je udio pojedinih glasova u grupi glasova (samo informativno).
    :param niz: Niz glasova.
    :return: Lista [(glas, postotak), ...]
    """
    return sorted(frekv(niz).items(), key=lambda x: x[1], reverse=True)


def izlazni_podaci(ms10, puls):
    """
    Ovo je pocetna funkcija iz koje se dobija rezultat prepoznavanja glasova
    od strane neuralne mreze. Ova funkcija grupira glasove prema zadanim
    kategorijama tako da se glasovi iste kategorije nalaze u jednom nizu koji
    predstavlja tu grupu glasova.

    :param ms10: Lista parova (glas, vrijeme)
    :param puls: Lista parova (glas, vrijeme)
    :return: Lista ntorki. FORMAT: (od, do, niz_glasova, udio_glasova)
    """
    # popis_glasova = objedini_popis_glasova(ms10, puls)
    popis_glasova = util.objedini(ms10, puls, lambda p: p[1])
    vremena = [a[1] for a in popis_glasova]
    slova = [a[0] for a in popis_glasova]
    popis = []
    prethodna_ntorka = None
    istaknuti_glasovi = re.finditer('([aeiou]+)|([šđžč]+)|([csz]+)',
                                    ''.join(slova))

    # dodaju se svi glasovi iz gornjeg RE (10 ms i pulsevi), a ostali glasovi
    # dolaze iz pulseva
    for glas in istaknuti_glasovi:
        string_glasova = glas.group()
        if len(string_glasova) < 3:  # ne gledamo nizove s manje od 3 glasa
            continue

        a = vremena[glas.span()[0]]
        b = vremena[glas.span()[1] - 1]
        udio_glasova = postotak_glasova(string_glasova)

        ntorka = (a, b, string_glasova, udio_glasova)

        # umetni ostale glasove koji se nalaze u ovom intervalu (ako postoje)
        if prethodna_ntorka is not None:
            pocetak_prethodni = prethodna_ntorka[0]
            pocetak_tekuci = ntorka[0]

            ostali_glasovi, od, do = glas_pulsa(puls, pocetak_prethodni,
                                                pocetak_tekuci)
            if len(ostali_glasovi) > 0:
                udio_ostalih_glasova = postotak_glasova(ostali_glasovi)
                popis += [(od, do, '*' + ostali_glasovi, udio_ostalih_glasova)]

        popis += [ntorka]
        prethodna_ntorka = ntorka

    return popis


def klasificiraj_glasove(direktorij):
    """
    Ovdje se klasificira cijela snimka.

    :return: Lista ntorki formata [(od, do, grupa glasova, statistika), ...]
    kao [(0.0894, 0.0947, 'eee', [('e', 100.0)]),
         (0.1, 0.1634, 'zzzzzszzzzz', [('z', 90.91), ('s', 9.09)]),
         ...]
    """
    klasf_ms10 = klasificiraj(util.konfig('datoteka', 'dir') + direktorij,
                              'ltas_10ms.txt',
                              'int_10ms.txt',
                              'klasifikacija_10ms.txt')

    klasf_puls = klasificiraj(util.konfig('datoteka', 'dir') + direktorij,
                              'ltas_puls.txt',
                              'int_puls.txt',
                              'klasifikacija_puls.txt')

    return izlazni_podaci(klasf_ms10, klasf_puls)


if __name__ == "__main__":
    treniraj()
