import statistics
from collections import OrderedDict
import configparser
import matplotlib.pyplot as plt
import numpy

import konstante


def slovo_glasa(oznaka):
    """ Konverzija nasih glasova u odgovarajuca slova. """

    mapa = {'sh': 'š',
            'ch': 'č',
            'zh': 'ž',
            'dzh': 'đ',
            'lj': '%',
            'nj': '#'}

    return mapa[oznaka] if oznaka in mapa else oznaka


def konfig(a, b):
    cfg = configparser.ConfigParser()
    cfg._interpolation = configparser.ExtendedInterpolation()
    cfg.read('konfig.ini')
    return cfg.get(a, b)


def graf(vrijednosti, oznake, x_labela, y_labela):
    """
    Crtanje bar-grafikona.
    :param vrijednosti:
    :param oznake: tekst ispod svakog stupica na x-koordinati
    :param x_labela:
    :param y_labela:
    :return: None
    """
    plt.figure(1, figsize=(10, 5), dpi=180, facecolor='white')
    plt.bar(numpy.arange(len(vrijednosti)), vrijednosti, color='gray')
    plt.xticks(numpy.arange(len(oznake)), oznake)
    plt.xlabel(x_labela)
    plt.ylabel(y_labela)
    ax = plt.axes()
    ax.yaxis.grid()
    plt.show()


class PripremaTeksta:
    zamjene = {
        ' ': '',
        'ts': 'c',
        'ds': 'c',
        'tz': 'c',
        'dz': 'z',
        'dp': 'tp',
        'tb': 'db',
        'dk': 'tk',
        'bk': 'pk',
        'tg': 'dg',
        'pg': 'bg',
        'dž': 'đ',
        'tć': 'č',
        'tč': 'č',
        'dć': 'č',
        'dč': 'č',
        'ć': 'č',
        'lj': '%',
        'nj': '#',
    }

    def __init__(self, tekst):
        self._obradjen_tekst = tekst
        self._pozicije = OrderedDict()

    def obradjen_tekst(self):
        for zamjena in PripremaTeksta.zamjene:
            p = self._obradjen_tekst.find(zamjena)
            while p >= 0:
                if zamjena in self._pozicije:
                    self._pozicije[zamjena] = \
                        (self._pozicije[zamjena][0],
                         self._pozicije[zamjena][1] + [p])
                else:
                    self._pozicije[zamjena] = \
                        (len(PripremaTeksta.zamjene[zamjena]), [p])

                self._obradjen_tekst = self._obradjen_tekst[:p] \
                                       + PripremaTeksta.zamjene[zamjena] \
                                       + self._obradjen_tekst[p + len(zamjena):]
                p = self._obradjen_tekst.find(zamjena)

        return self._obradjen_tekst

    def izvorni_niz(self, niz):
        """
        Originalni tekst se vraca zamjenama u suprotnom redoslijedu od onoga
        kojim su u originalnom tekstu napravljene zamjene.

        :return: Originalni tekst.
        """
        # vrijednost kljuca 'za_umetnuti' je par (duljina, [pozicije...])
        for za_umetnuti in reversed(self._pozicije):
            info = self._pozicije[za_umetnuti]
            duljina, pozicije = info
            for p in reversed(pozicije):
                niz = niz[:p] + list(za_umetnuti) + niz[p + duljina:]

        return niz


def prikazi_tabelu_glasova(tabela):
    zaglavlje = f'{"R.BR.":7}{"OD":10}{"DO":10}{"DULJINA":10}{"GLASOVI":70}'
    crta = f'{"-----":7}{"--":10}{"--":10}{"-------":10}{"-------":70}'
    print(zaglavlje)
    print(crta)

    rbr = 0
    for redak in tabela:
        od = redak['od']
        do = redak['do']
        glasovi = redak['grupa_glasova']
        formatirano = f'{str(rbr) + ":":<7}' \
                      f'{round(od, 2):<10}' \
                      f'{round(do, 2):<10}' \
                      f'{round((do - od) * 1000, 2):<10}' \
                      f'{glasovi:70}'
                      # f'{str(redak["statistika"]):40}'
        print(formatirano)
        rbr += 1


def prikazi_tabelu_rijeci(tabela, ime_datoteke=None):
    """
    Prikaz podataka prepoznatog govora.

    :param tabela: Tabela s podacima za svaku rijec.
    :param ime_datoteke:
    :return: None
    """
    # ispisi rijeci s vremenima i vremenskim odstupanjima
    print(f'{"RIJEC":20}{"OD":8}{"DO":8}'
          f'{"ODSTUPANJE-OD":>14}{"ODSTUPANJE-DO":>14}'
          f'{"SLOGOVA":^11}'
          f'{"INTENZITET (MIN/MAKS/PR)":<30}'
          f'{"TON (MIN/MAKS/PR)":<30}'
          f'{"TRAJANJE":^12}'
          f'{"ISTAKNUTO (I/T/R)":<15}')

    datoteka = None
    if ime_datoteke is not None:
        datoteka = open(ime_datoteke, 'w', encoding='utf-8')

    odstupanja = []
    for redak in tabela:
        rijec = redak['rijec']
        od = str(round(redak["od"], 2))
        do = str(round(redak["do"], 2))
        # ispravno_od = str(redak["ispravno_od"])
        # ispravno_do = str(redak["ispravno_do"])
        odstupanje_od = str(round(redak['odstupanje_od'], 2))
        odstupanje_do = str(round(redak['odstupanje_do'], 2))
        odstupanja += [(abs(round(redak['odstupanje_od'], 2)),
                        abs(round(redak['odstupanje_do'], 2)))]
        br_slogova = str(len(redak["slogovi"]))
        # slogovi = '-'.join(redak["slogovi"])

        int_ispis = f'{str(redak["intenzitet"][0]):8}' \
                    f'{str(redak["intenzitet"][1]):8}' \
                    f'{str(redak["intenzitet"][2]):8}'
        ton_ispis = f'{str(redak["ton"][0]):8}' \
                    f'{str(redak["ton"][1]):8}' \
                    f'{str(redak["ton"][2]):8}'

        istaknuto = ('I' if redak['istaknut_intenzitet'] else '') \
                    + ('T' if redak['istaknut_ton'] else '') \
                    + ('R' if redak['istaknuto_trajanje'] else '')

        tekst = f'{rijec:20}' \
                f'{od:8}' \
                f'{do:8}' \
                f'{odstupanje_od:>14}' \
                f'{odstupanje_do:>14}' \
                f'{br_slogova:^11}' \
                f'{int_ispis:^30}' \
                f'{ton_ispis:^30}' \
                f'{str(redak["trajanje"]):^12}' \
                f'{istaknuto:<15}'
        print(tekst)

        if datoteka is not None:
            datoteka.write(
                f'{rijec},'
                f'{od},'
                f'{do},'
                f'{odstupanje_od},'
                f'{odstupanje_do},'
                f'{br_slogova},'
                f'{redak["intenzitet"][0]}/{redak["intenzitet"][1]}'
                f'/{redak["intenzitet"][2]},'
                f'{redak["ton"][0]}/{redak["ton"][1]}/{redak["ton"][2]},'
                f'{str(redak["trajanje"])},'
                f'{istaknuto}' +
                '\n')

    if datoteka is not None:
        datoteka.close()

    prosj_trajanje_glasa = int(statistics.mean(
        konstante.PROSJ_TRAJANJE_GLASOVA.values()))

    print()
    print(f'{"UKUPAN BROJ RIJEČI:":37}{len(tabela):<10}')

    maks_odstup_od = max(list(map(lambda p: p[0], odstupanja)))
    maks_odstup_do = max(list(map(lambda p: p[1], odstupanja)))
    print(f'{"MAKS. ODSTUPANJE (ms):":37}'
          f'{str((round(maks_odstup_od, 2), round(maks_odstup_do, 2))):10}')

    min_odstup_od = min(list(map(lambda p: p[0], odstupanja)))
    min_odstup_do = min(list(map(lambda p: p[1], odstupanja)))
    print(f'{"MIN. ODSTUPANJE (ms):":37}'
          f'{str((round(min_odstup_od, 2), round(min_odstup_do, 2))):10}')

    prosj_odstup_od = statistics.mean(list(map(lambda p: p[0], odstupanja)))
    prosj_odstup_do = statistics.mean(list(map(lambda p: p[1], odstupanja)))

    gl_od = round(prosj_odstup_od / prosj_trajanje_glasa, 1)
    gl_do = round(prosj_odstup_do / prosj_trajanje_glasa, 1)
    gl_prosjek = round((gl_od + gl_do) / 2, 1)
    print(f'{"PROSJEČNO ODSTUPANJE (br. glasova):":37}'
          f'{str((gl_od, gl_do)) + " = " + str(gl_prosjek):10}')

    print(f'{"PROSJEČNO ODSTUPANJE (ms):":37}'
          f'{str((round(prosj_odstup_od, 2), round(prosj_odstup_do, 2))):10}')

    # ispis tablice odstupanja do N milisekundi

    # prosjecno trajanje glasa je oko 60 ms
    filter_odstupanja = (prosj_trajanje_glasa, prosj_trajanje_glasa * 2,
                         prosj_trajanje_glasa * 3, prosj_trajanje_glasa * 4,
                         prosj_trajanje_glasa * 5)

    print(f'{"ODST.(" + str(prosj_trajanje_glasa) + "ms/gl.)":18}'
          f'{"BROJ-RIJEČI":13}{"%":8}')

    for odst in filter_odstupanja:
        popis = list(filter(lambda p: abs(p[0]) <= odst and abs(p[1]) <= odst,
                            odstupanja))
        print(f'{int(odst / prosj_trajanje_glasa):<18}'
              f'{len(popis):<13}'
              f'{round(len(popis) / len(tabela) * 100, 2):<8}')


def prikazi_tempo(grupe, prikazi_graf=False):
    print(f'{"INTERVAL":18}{"BR_SLOGOVA":12}{"TRAJANJE":10}')
    trajanja = []
    for grupa in grupe:
        od = round(grupa[0]['od'], 2)
        do = round(grupa[-1]['do'], 2)

        br_slogova = 0
        for e in grupa:
            br_slogova += e['broj_slogova']

        duljina = do - od
        trajanje_grupe = round(br_slogova / duljina, 2)
        print(f'{str((od, do)):<18}{br_slogova:^12}'
              f'{trajanje_grupe:^10}')
        trajanja += [trajanje_grupe]

    # graf_analiza_tempa(trajanja)
    if prikazi_graf:
        graf(trajanja, trajanja, 'syllable group duration (milliseconds)',
             'duration (milliseconds)')


def ucitaj_titlove(direktorij):
    """
    Ovdje se podrazumijeva da su titlovi prethodno spremljeni u tekstualnu
    datoteku (nakon što su učitani iz .ts datoteke).

    :param direktorij:
    :return:
    """
    try:
        with open(konfig('datoteka', 'dir')
                  + direktorij
                  + r'\titlovi.txt', 'r', encoding='utf8') as f:
            redovi = f.readlines()

        titlovi = []
        for red in redovi:
            navodnik = red.find('"')
            interval = red[:navodnik - 1].split(' ')
            tekst = red[navodnik + 1:-2]  # bez drugog navodnika
            titlovi += [(float(interval[0]), float(interval[1]), tekst)]

        return titlovi
    except Exception as e:
        print('??? UCITAVANJE TITLOVA:', e)


def objedini(niz_a, niz_b, fn_pozicija):
    """
    Spajanje popisa za 10ms i pulseve.

    :param niz_a:
    :param niz_b:
    :param fn_pozicija: Funkcija koja za zadanu ntorku daje vrijednost za
                        usporedbu.
    :return:
    """
    indeks_a = 0
    indeks_b = 0
    rezultat = []
    while indeks_a < len(niz_a) or indeks_b < len(niz_b):
        if indeks_a < len(niz_a):
            vr_a = niz_a[indeks_a]
        else:
            rezultat.extend(niz_b[indeks_b:])  # dodaj sto je ostalo
            return rezultat

        if indeks_b < len(niz_b):
            vr_b = niz_b[indeks_b]
        else:
            rezultat.extend(niz_a[indeks_a:])  # dodaj sto je ostalo
            return rezultat

        a = fn_pozicija(vr_a)
        b = fn_pozicija(vr_b)
        if a < b or indeks_b >= len(niz_b):
            rezultat += [vr_a]
            indeks_a += 1
        elif b < a or indeks_a >= len(niz_a):
            rezultat += [vr_b]
            indeks_b += 1
        else:
            rezultat += [vr_a, vr_b]
            indeks_a += 1
            indeks_b += 1

    return rezultat


def ucitaj_prozodijske_podatke(direktorij, *datoteke):
    """
    Ulazni podaci moraju biti formata v1:v2, npr. 2.177:75.94883

    :param direktorij:
    :type datoteke: Niz datoteka cije podatke treba objediniti
    :return: Lista parova (v1, v2)
    """
    podaci = []
    for dat in datoteke:
        vrijednosti = []
        for red in open(konfig('datoteka', 'dir') + direktorij + '\\' + dat):
            try:
                vrijeme_str, intenzitet_str = red.split(':')
                vrijeme = float(vrijeme_str)
                intenzitet = float(intenzitet_str)
                vrijednosti += [(vrijeme, intenzitet)]
            except Exception:
                pass  # preskacemo nepotpune podatke

        podaci += [vrijednosti]
        if len(podaci) == 2:  # ako imamo dvije liste, objedini ih
            podaci = objedini(podaci[0], podaci[1], lambda p: p[0])

    # print(podaci)
    return podaci


def sekunde(minute, sekunde):
    return minute * 60 + sekunde


if __name__ == '__main__':
    pass
